{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "gHoQThpObPSq"
      },
      "source": [
        "# 第四次作业\n",
        "\n",
        "在本次作业中，大家需要实现SIGN模型并在cora数据集上进行测试。\n",
        "\n",
        "本作业需要安装[CogDL](https://github.com/THUDM/cogdl)：pip install cogdl\n",
        "\n",
        "如需使用gpu版，请先安装gpu版本的[PyTorch](https://pytorch.org/get-started/locally/)，再安装cogdl。\n",
        "\n",
        "本作业由智谱GNN中心及课程团队筹备，由CogDL团队提供技术支持。\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### 1. 实现SIGN模型\n",
        "可以调用CogDL提供的spmm(graph, x)函数来进行稀疏矩阵乘AX。\n",
        "模型实现可以参考原SIGN论文：https://arxiv.org/pdf/2004.11198.pdf 。"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {},
      "outputs": [],
      "source": [
        "import os\n",
        "import torch\n",
        "import torch.nn as nn\n",
        "import torch.nn.functional as F\n",
        "\n",
        "from cogdl.models.nn.mlp import MLP\n",
        "from cogdl.utils import spmm\n",
        "\n",
        "\n",
        "class SIGN(nn.Module):\n",
        "\n",
        "    def __init__(\n",
        "        self,\n",
        "        num_features,\n",
        "        hidden_size,\n",
        "        num_classes,\n",
        "        num_layers,\n",
        "        dropout,\n",
        "        nhop,\n",
        "        norm=\"batchnorm\",\n",
        "        activation=\"relu\",\n",
        "    ):\n",
        "        super(SIGN, self).__init__()\n",
        "        self.nhop = nhop\n",
        "        num_features = num_features * (1 + nhop) # nhop: number of hops\n",
        "        self.mlp = MLP(\n",
        "            in_feats=num_features,\n",
        "            out_feats=num_classes,\n",
        "            hidden_size=hidden_size,\n",
        "            num_layers=num_layers,\n",
        "            dropout=dropout,\n",
        "            activation=activation,\n",
        "            norm=norm,\n",
        "        )\n",
        "\n",
        "        self.all_x = None\n",
        "\n",
        "    def preprocessing(self, graph, x):\n",
        "        graph.sym_norm() # 对图进行对称归一化\n",
        "\n",
        "        ###################\n",
        "        ##### 作业填空 #####\n",
        "        ###################\n",
        "        # 进行多次feature propagation，并将所有特征拼接起来，特征维度与MLP模型输入一致\n",
        "    \n",
        "        self.all_x = ...\n",
        "\n",
        "    def forward(self, graph):\n",
        "        ###################\n",
        "        ##### 作业填空 #####\n",
        "        ###################\n",
        "        # 将MLP模型应用到预先计算出来的特征中\n",
        "\n",
        "        ret = ...\n",
        "        return ret"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "OAPbTveYjONP"
      },
      "source": [
        "### 2. 从cogdl中加载cora数据集（x表示特征，y表示标签，mask表示训练/验证/测试集的划分）"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 2,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "61aEVO2ySq27",
        "outputId": "6da332b4-d5aa-4baf-d95e-4dd5b4b291da"
      },
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Graph(x=[2708, 1433], y=[2708], train_mask=[2708], val_mask=[2708], test_mask=[2708], edge_index=[2, 10556])\n"
          ]
        }
      ],
      "source": [
        "from cogdl.datasets import build_dataset_from_name\n",
        "\n",
        "dataset = build_dataset_from_name(\"cora\")\n",
        "data = dataset[0]\n",
        "print(data)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "FcRyHUuIjhfT"
      },
      "source": [
        "### 3. 使用你实现的SIGN模型进行训练"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 3,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "pr5lZC2eWCYm",
        "outputId": "c42b59ef-9637-4bac-e1f1-79a3f9f6228a"
      },
      "outputs": [
        {
          "name": "stderr",
          "output_type": "stream",
          "text": [
            "Epoch: 099, Train Loss: 0.0002, Val Loss: 0.9386, Val Acc: 0.7360: 100%|██████████| 100/100 [00:15<00:00,  6.28it/s]\n"
          ]
        },
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Val Acc 0.74\n",
            "Test Acc 0.767\n"
          ]
        }
      ],
      "source": [
        "import math\n",
        "import copy\n",
        "from tqdm import tqdm\n",
        "\n",
        "def accuracy(y_pred, y_true):\n",
        "    y_true = y_true.squeeze().long()\n",
        "    preds = y_pred.max(1)[1].type_as(y_true)\n",
        "    correct = preds.eq(y_true).double()\n",
        "    correct = correct.sum().item()\n",
        "    return correct / len(y_true)\n",
        "\n",
        "def train(hidden_size=128, nhop=3):\n",
        "    hidden_size = hidden_size\n",
        "    model = SIGN(data.x.shape[1], hidden_size, data.y.max() + 1, num_layers=2, dropout=0.5, nhop=nhop)\n",
        "    model.preprocessing(data, data.x)\n",
        "\n",
        "    if torch.cuda.is_available():\n",
        "        device = torch.device(\"cuda\")\n",
        "        model = model.to(device)\n",
        "        data.to(device)\n",
        "\n",
        "    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "    epoch_iter = tqdm(range(100), position=0, leave=True)\n",
        "    best_model = None\n",
        "    best_loss = 1e8\n",
        "    for epoch in epoch_iter:\n",
        "        model.train()\n",
        "        optimizer.zero_grad()\n",
        "        logits = model(data)\n",
        "        loss = F.cross_entropy(logits[data.train_mask], data.y[data.train_mask])\n",
        "        loss.backward()\n",
        "        optimizer.step()\n",
        "        train_loss = loss.item()\n",
        "\n",
        "        model.eval()\n",
        "        with torch.no_grad():\n",
        "            logits = model(data)\n",
        "            val_loss = F.cross_entropy(logits[data.val_mask], data.y[data.val_mask]).item()\n",
        "            val_acc = accuracy(logits[data.val_mask], data.y[data.val_mask])\n",
        "            if val_loss < best_loss:\n",
        "                best_loss = val_loss\n",
        "                best_model = copy.deepcopy(model)\n",
        "\n",
        "        epoch_iter.set_description(f\"Epoch: {epoch:03d}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}, Val Acc: {val_acc:.4f}\")\n",
        "\n",
        "    with torch.no_grad():\n",
        "        logits = best_model(data)\n",
        "        val_acc = accuracy(logits[data.val_mask], data.y[data.val_mask])\n",
        "        test_acc = accuracy(logits[data.test_mask], data.y[data.test_mask])\n",
        "    print(\"Val Acc\", val_acc)\n",
        "    print(\"Test Acc\", test_acc)\n",
        "\n",
        "train()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### 4. 设置不同的hidden_size和nhop数进行实验，分析这两个参数的影响\n",
        "\n",
        "hidden_size可以从64, 128, 256, 512中选取，nhop可以从2～6中选取"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 4,
      "metadata": {},
      "outputs": [
        {
          "name": "stderr",
          "output_type": "stream",
          "text": [
            "Epoch: 099, Train Loss: 0.0034, Val Loss: 1.0779, Val Acc: 0.6620: 100%|██████████| 100/100 [00:06<00:00, 14.61it/s]\n"
          ]
        },
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Val Acc 0.67\n",
            "Test Acc 0.679\n"
          ]
        }
      ],
      "source": [
        "train(hidden_size=64, nhop=2)"
      ]
    }
  ],
  "metadata": {
    "accelerator": "GPU",
    "colab": {
      "collapsed_sections": [],
      "name": "write-your-first-gcn",
      "provenance": []
    },
    "kernelspec": {
      "display_name": "Python 3.6.15 ('py3.6')",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.15"
    },
    "vscode": {
      "interpreter": {
        "hash": "7703d8cc28a94efd96474faf641471b2ac55415e3c97cc8df6cd7b51cf78a952"
      }
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}
